/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is FullerScreen.
 *
 * The Initial Developer of the Original Code is
 * Olivier Gambier
 * Portions created by the Initial Developer are Copyright (C) 2006-2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  DISRUPTIVE INNOVATIONS SARL
 *
 *
 * ***** END LICENSE BLOCK ***** */

/**
 * Cleaned-up and packaged from version 0.1.0 (2006/04/10)
 */

// src/javascript.js

/**
 * This function adapted from http://dean.edwards.name
 * Mmm. I pushed ^ and $ in there. Wonder why Dean didn't do it...
 */

function gEBS_regEscape(string) {
	var reg = new RegExp("([\\^\\$\\/()[\\]?{}|*+-])", 'g');
	return string.replace(reg, "\\$1");
//	return string.replace(/([\^\$\/()[\]?{}|*+-])/g, "\\$1");
}

/**
 * A simple "contains" extension to the Array prototype.
 */

function gEBS_arrayContains(aArray, aThing)
{
    for (var x = 0; x < aArray.length; x++){
        if(aArray[x] == aThing){
            return true;
        }
    }
    return false;
};

/**
 * Returns the key. Like array_search does in php.
 */

/**
 * Wander through the DOM.
 * A classic from IBM (author: Dethe Elza).
 */

function gEBS_nextNode(node){
    if (!node){return null;}
    if (node.firstChild){
        return node.firstChild;
    }else{
        return gEBS_nextWide(node);
    }
};

function gEBS_nextWide(node){
    if (!node) {return null;}
    if (node.nextSibling){
        return node.nextSibling;
    }else{
        return gEBS_nextWide(node.parentNode);
	}
};

/**
 * Error handler
 */

function gEBS_printfire()
{
// Do whatever you want with arguments
}


// src/handlers/core.js",

/**
 * Beside actual handlers, we need to retrieve the list at load time, to
 * actually build up the list.
 */

var gEBS_cssHandlers = {
	onload: function(type){
		var result = [];
		for(var x in this[type]){
			if(x.charAt(0) != '_'){
			// May be externalized: css2js
				var cssNameString = '';
				for(var y = 0; y < x.length; y++){
					if(x.charAt(y).toLowerCase() != x.charAt(y)){
						cssNameString += '-' + x.charAt(y).toLowerCase();
					} else {
						cssNameString += x.charAt(y);
					}
				}
				result.push(gEBS_regEscape(cssNameString));
			}
		}
		return '(?:'+result.join('|')+')';
	}
};

gEBS_cssHandlers.combinators = {};
gEBS_cssHandlers.matchers = {};
gEBS_cssHandlers.properties = {};
gEBS_cssHandlers.pseudoClasses = {};
gEBS_cssHandlers.pseudoElements = {};



// src/handlers/combinators/level3.js",


/**
 * Generic handler to call whenever we need to "combine"
 * wayToGo should be either "up" or "down", and of course it MUST exist
 */

gEBS_cssHandlers.combinatorsHandler = function(element, combinator, wayToGo){
	return this.combinators[combinator][wayToGo](element);
};

/**
 * CSS level 3 combinators
 */

gEBS_cssHandlers.combinators['+'] = {
	up: function(element){
		var candidates = [];
	    var previous = element.previousSibling;
	    while(previous && (previous.nodeType != 1))
	        previous = previous.previousSibling;
	    if(previous && (previous.nodeType == 1))
	        candidates = Array(previous);
		return candidates;
	},
	down: function(element){
		var candidates = [];
	    var next = element.nextSibling;
	    while(next && (next.nodeType != 1))
	        next = next.nextSibling;
	    if(next && (next.nodeType == 1))
	        candidates = Array(next);
		return candidates;
	}
};

gEBS_cssHandlers.combinators['>'] = {
	up: function(element){
	    if(element.parentNode.nodeType == 1)
	        return [element.parentNode];
		else
			return [];
	},
	down: function(element){
		var candidates = [];
	    var next = element.firstChild;
	    while(next){
	        if(next.nodeType == 1)
	            candidates.push(next);
	        next = next.nextSibling;
	    }
		return candidates;
	}
};

gEBS_cssHandlers.combinators['~'] = {
	up: function(element){
		var candidates = [];
	    var previous = element.previousSibling;
	    while(previous){
	        if(previous.nodeType == 1)
	            candidates.push(previous);
	        previous = previous.previousSibling;
	    }
		return candidates;
	},
	down: function(element){
		var candidates = [];
	    var next = element.nextSibling;
	    while(next){
	        if(next.nodeType == 1)
	            candidates.push(next);
	        next = next.nextSibling;
	    }
		return candidates;
	}
};

gEBS_cssHandlers.combinators[' '] = {
	up: function(element){
		var candidates = [];
	    var parent = element.parentNode;
	    while(parent && parent.nodeType == 1){
	        candidates.push(parent);
	        parent = parent.parentNode;
	    }
	    return candidates;
	},
	down: function(element){
// Check that actually works better, or even work at all!!!
	    return element.getElementsByTagNameNS('*', '*');
	}
};



// src/handlers/matchers/level3.js",


/**
 * Generic handler to call whenever we need to check something
 */

gEBS_cssHandlers.matchersHandler = function(valueToMatch, matchingMethod, realContent){
	return this.matchers[matchingMethod](valueToMatch, realContent);
};


/**
 * Matching methods defined by css l3
 */

gEBS_cssHandlers.matchers['='] = function(valueToMatch, realContent){
	return ((realContent == valueToMatch) ? true : false);
};

gEBS_cssHandlers.matchers['^='] = function(valueToMatch, realContent){
	return (realContent.match('^'+gEBS_regEscape(valueToMatch)) ? true : false);
};

gEBS_cssHandlers.matchers['$='] = function(valueToMatch, realContent){
	return (realContent.match(gEBS_regEscape(valueToMatch)+'$') ? true : false);
};

gEBS_cssHandlers.matchers['*='] = function(valueToMatch, realContent){
	return (realContent.match(gEBS_regEscape(valueToMatch)) ? true : false);
};

gEBS_cssHandlers.matchers['~='] = function(valueToMatch, realContent){
	var exit = !valueToMatch.match(/ /);
	var reg = new RegExp('(?:^|\\s)'+gEBS_regEscape(valueToMatch)+'(?:$|\\s)');
	var exit2 = realContent.match(reg);
	return ((exit && exit2) ? true : false);
};

gEBS_cssHandlers.matchers['|='] = function(valueToMatch, realContent){
	return (realContent.match('^'+gEBS_regEscape(valueToMatch)+'(?:-|$)') ? true : false);
};



// src/handlers/pseudoElements/level3.js",

/**
 * Generic handler for pseudo elements. Still need to figure out what to do.
 */

gEBS_cssHandlers.pseudoElementsHandler = function(element, pseudoName){
	return this.pseudoElements[pseudoName](element);
};

gEBS_cssHandlers.pseudoElements.firstLine = function(element){return element;};
gEBS_cssHandlers.pseudoElements.before = function(element){return element;};
gEBS_cssHandlers.pseudoElements.after = function(element){return element;};
gEBS_cssHandlers.pseudoElements.firstLetter = function(element){return element;};


// src/handlers/pseudoClasses/level3.js",

/**
 * Generic handler
 */

gEBS_cssHandlers.pseudoClassesHandler = function(element, pseudoName, optionnalArguments){
	return this.pseudoClasses[pseudoName](element, optionnalArguments);
};

/**
 * CSS Level 3 Pseudos (http://www.w3.org/TR/css3-selectors/)
 */


/** 
 * 6.6.1. Dynamic pseudo-classes
 * They do always return "true". Dynamic styles can't be dealt with on that level -
 * as we don't have any knowledge about what will be done will the selector.
 * A proper event manager should be implemented in the doubleCross lib, or elsewhere,
 * but not here.
 * Still, we need to recognize these pseudos, so "true".
 * The link pseudo-classes: :link and :visited
 * The user action pseudo-classes :hover, :active, and :focus
 */


gEBS_cssHandlers.pseudoClasses.focus = function(element){
	return true;
};

gEBS_cssHandlers.pseudoClasses.active = function(element){
	return true;
};

gEBS_cssHandlers.pseudoClasses.hover = function(element){
	return true;
};

// FIXME
// Visited should apply on visited links.
// How am I supposed to know that? Play with history????
// Anyhow, what does :link means outside of html???????
// Or: maybe I'm supposed to... deal links my way, without asking the bro????
// Try something with links????
gEBS_cssHandlers.pseudoClasses.link = function(element){
	return (gEBS_cssCheckers.nameAndNS(element, '*', 'a') || gEBS_cssCheckers.nameAndNS(element, '*', 'link'));
};

gEBS_cssHandlers.pseudoClasses.visited = function(element){
	return (gEBS_cssCheckers.nameAndNS(element, '*', 'a') || gEBS_cssCheckers.nameAndNS(element, '*', 'link'));
};

/**
 * 6.6.4. UI element states pseudo-classes: :checked, :enabled, :disabled, :indeterminate
 */

gEBS_cssHandlers.pseudoClasses.checked = function(element){
	return (gEBS_cssCheckers.attribute(element, '*', 'checked') || gEBS_cssCheckers.attribute(element, '*', 'selected'));
};

// Probably just wrong...
gEBS_cssHandlers.pseudoClasses.enabled = function(element){
	return (gEBS_cssCheckers.attribute(element, '*', 'enabled') || !gEBS_cssCheckers.attribute(element, '*', 'disabled'));
};

// Probably just wrong...
gEBS_cssHandlers.pseudoClasses.disabled = function(element){
	return gEBS_cssCheckers.attribute(element, '*', 'disabled');
};




/**
 * 6.6.2. The :target pseudo-class
 * FIXME: How is it supposed to work with xml:id?
 */

gEBS_cssHandlers.pseudoClasses.target = function(element){
	return gEBS_cssCheckers.attribute(element, '*', 'id', '=', window.location.hash.substr(1));
};


/**
 * 6.6.3. The :lang() pseudo-class
 * MAY fixme: in html4, meta is supposed to influence the lang
 */

gEBS_cssHandlers.pseudoClasses.lang = function(element, langString){
	if(langString){
		var exit = false;
		var upStair = element;
		while(!exit && (upStair.nodeType == 1)){
			if (gEBS_cssCheckers.attribute(element, '*', 'lang', '|=', langString))
				exit = true;
			upStair = upStair.parentNode;
		}
// Lang cannot be null
	}else{
		var exit = -1;
	}
	return exit;
};



/**
 * 6.6.5. Structural pseudo-classes
 */

/**
 * o :root pseudo-class
 */

gEBS_cssHandlers.pseudoClasses.root = function(element){
	if (element.parentNode.nodeType == 1){
		return false
	}else{
		return true;
	}
};
          
/**
 * o :nth-child() pseudo-class
 * o :nth-last-child()
 * o :first-child pseudo-class
 * o :last-child pseudo-class
 */

gEBS_cssHandlers.pseudoClasses.nthChild = function(element, fragment){
	return this.____cssPseudoNth(element, fragment, false, false);
};

gEBS_cssHandlers.pseudoClasses.nthLastChild = function(element, fragment){
	return this.____cssPseudoNth(element, fragment, true, false);
};

gEBS_cssHandlers.pseudoClasses.firstChild = function(element){
	return this.____cssPseudoNth(element, '1', false, false);
};

gEBS_cssHandlers.pseudoClasses.lastChild = function(element){
	return this.____cssPseudoNth(element, '1', true, false);
};

/**
 * o :nth-of-type() pseudo-class
 * o :nth-last-of-type()
 * o :first-of-type pseudo-class
 * o :last-of-type pseudo-class
 */

gEBS_cssHandlers.pseudoClasses.nthOfType = function(element, fragment){
	return this.____cssPseudoNth(element, fragment, false, true);
};

gEBS_cssHandlers.pseudoClasses.nthLastOfType = function(element, fragment){
	return this.____cssPseudoNth(element, fragment, true, true);
};

gEBS_cssHandlers.pseudoClasses.firstOfType = function(element){
	return this.____cssPseudoNth(element, '1', false, true);
};

gEBS_cssHandlers.pseudoClasses.lastOfType = function(element){
	return this.____cssPseudoNth(element, '1', true, true);
};

/**
 * o :only-child pseudo-class
 * o :only-of-type pseudo-class
 * o :empty pseudo-class
 */

gEBS_cssHandlers.pseudoClasses.onlyChild = function(element, fragment){
	return (this.lastChild(element) && this.firstChild(element));
};

gEBS_cssHandlers.pseudoClasses.onlyOfType = function(element){
	return (this.lastOfType(element) && this.firstOfType(element));
};

gEBS_cssHandlers.pseudoClasses.empty = function(element){
	var next = element.firstChild;
    var exit = true;
	while(exit && next){
		if(next.nodeType == 1 || next.nodeType == 3)
            exit = false;
		next = next.nextSibling;
	}
    return exit;
};

/**
 * 6.6.7. The negation pseudo-class
 */

gEBS_cssHandlers.pseudoClasses.not = function(element, fragment){
// fragment is supposed already resolved
// anyhow, I'm not following w3c on this one, as fragment may be more complex
// than what they allow (including nested :not)
// problem2 - how to deal with dynamic pseudos?
	return !(gEBS_cssCheckers.sequence(element, fragment));
};


// Lil helper for all the position pseudos testers
gEBS_cssHandlers.pseudoClasses.____cssPseudoNth = function(element, formula, last, type){
// NOTE: all these suppose we GOT a parent.
	var found = false;
	if(element.parentNode.nodeType == 1){
	// FIXME: capitalization on nodeNames and stoopid HTML?
	    var method = (last ? 'nextSibling' : 'previousSibling');
	// Where are we?
	    var amAt = 1;
	    var next = element[method];
	    while(next){
	        if(next.nodeType == 1 && (!type || (next.nodeName.toLowerCase() == element.nodeName.toLowerCase())))
	            amAt++;
	        next = next[method];
	    }
		if(formula == 'odd')
			formula = '2n+1';
		if(formula == 'even')
			formula = '2n';
		// FIXME. This is *absolutely* awfull. Use modulos darling. Or split the stack sooner.
//		formula = formula.replace(/([0-9])([^\*\/\+\-0-9])/, "$1*$2");
		var reg = new RegExp("([0-9])([^\\*\\/\\+\\-0-9])");
		formula = formula.replace(reg, "$1*$2");
		var reg = new RegExp("([^\\*\\/\\+\\-0-9])");
		var n = 0;
		while(n < 100 && !found){
//			if (eval(formula.replace(/([^\*\/\+\-0-9])/, n)) == amAt)
			if (eval(formula.replace(reg, n)) == amAt)
				found = true;
			n++;
		}
	}
	return found;
};

// src/parser/core.js",


/**
 * The selector parser.
 * "group" and "sequence" may be publicly accessed
 */

var gEBS_cssSelectorParser = {};

/**
 * A generic function to map js names to css ones
 */

function gEBS_mapCssName2Js(cssNameString){
// Css names are supposed case insensitive.
	cssNameString = cssNameString.toLowerCase();
	var jsNameString = '';
	for(var x = 0; x < cssNameString.length; x++){
		if(cssNameString.charAt(x) == '-'){
			x++;
			jsNameString += cssNameString.charAt(x).toUpperCase();
		}else{
			jsNameString += cssNameString.charAt(x);
		}
	}
	return jsNameString;
};


/**
 * Publicly available thingies
 */

/**
 * (function) gEBS_cssSelectorParser.addNamespace(uri, prefix)
 * allow one to add a namespace. If no prefix is provided, it will be the default namespace.
 */

/**
 * (function) gEBS_cssSelectorParser.clearNameSpaces(prefix, uri)
 * reset namespaces...
 */

/**
 * (function) gEBS_cssSelectorParser.loadGroup(selectorGroup)
 * will parse a group, returning unparsable data and setting the group representation
 * in gEBS_cssSelectorParser.group
 */

/**
 * (array) gEBS_cssSelectorParser.group
 *
 * Will hold a representation of the last parsed group
 * eg: Array(
 * 		0 => (array) selectorRepresentation,
 * 		...,
 * 		n => (array) selectorRepresentation
 * )
 *
 * Each selectorRepresentation is:
 * Array(
 * 		0 => (object) sequenceObject,
 * 		1 => (string) combinator,
 * 		2n => (object) sequenceObject,
 * 		2n+1 => (string) combinator,
 * 		...
 *		m => (int) selector weight
 * )
 * 
 * A sequenceObject will hold
 *		selected: false,
 *		tagName: '*',
 *		nameSpace: '*',
 *		ids: [],
 *		classes: [name, ...]
 *		attributes: [name, [relation, value]],
 *		pseudoClasses: [name, [argumentStack]],
 *		pseudoElements: [name],
 *		weight: 0
 *
 */


/**
 * MUST be called onLoad to build-up some regexp depending on custom adds-ons
 */

gEBS_cssSelectorParser.onload = function(){
// Some implementation dependent data types
	var combinators = gEBS_cssHandlers.onload('combinators');
	var matchPattern = gEBS_cssHandlers.onload('matchers');
	var pseudoList = ':'+gEBS_cssHandlers.onload('pseudoElements') + '|' + gEBS_cssHandlers.onload('pseudoClasses');
// Generic css3 data types
	var __string = '(?:"(?:[^"]*)?"|\'(?:[^\']*)?\')';
	var __nonascii = '[^\\x00-\\x7f]';


//	var __nmstart = '(?:[_a-z]|'+__nonascii+')';
//	var __nmchar = '(?:[_a-z0-9-]|'+__nonascii+')';

// FIXME: this is dumb
	var __escape = '\\\\[^\\n\\r\\f0-9a-z]';
	var __nmstart = '(?:[_a-z]|'+__nonascii+'|'+__escape+')';
	var __nmchar = '(?:[_a-z0-9-]|'+__nonascii+'|'+__escape+')';


	var __ident = '[-]?'+__nmstart+__nmchar+'*';
	var __namespacePrefix = '(?:'+__ident+'|\\*)?\\|';
	var __number = '(?:[0-9]+|[0-9]*\.[0-9]+)';
	var __attributeName = '(?:'+__namespacePrefix+')?(?:'+__ident+')';
// Note that the spec sux here, by allowing only ident and strings as args
// We make it swallow numbers also
	var __attributeValue = '(?:'+__number+'|'+__ident+'|'+__string+')';

// Private extension: the "formula" type. Maybe extend this a bit.
// unicode 945-969: greek letters
// Ideas: deal with complex parenthesis
// Ideas: use MathML unicode range
	var __letter = '[a-z\\u0945-\\u0969]';
	var __atom = '(?:'+__number + __letter + '?|'+__letter+')';
	var __operations = '[\\+=\\-\\/\\*\\^]';
	var __formula = '(?:\\+|\\-)?' + __atom + '(?:\\s*'+ __operations + '\\s*' + __atom +')*';

// ########  Type selectors, honoring namespaces prefix.
	var typeSelector = '(?:'+__namespacePrefix+')?(?:'+__ident+'|\\*)';
// ######## Class, nothing to see here. Move on people.
	var cssClass = '\\.('+__ident+')';
// ########  Hash
	var id = '#('+__nmchar+'+)';
// ######## CSS3 Attributes
	var attribute = '\\[\\s*('+__attributeName+')\\s*(?:('+matchPattern+')\\s*('+__attributeValue+'))?\\s*\\]';

	var __atomicExpression = '(?:'+matchPattern+'|'+__string+'|'+__formula+'|'+__ident+')';
	var __argStack = __atomicExpression+'(?:\\s*,\\s*'+__atomicExpression+')*';
// Note: pseudoList should be __ident
	var __simplePseudoFunctionnal = ':(:?'+pseudoList+')(?:\\(\\s*('+__argStack+')?\\s*\\))?';
// This is for "not" and maybe others, that may take complex arguments	
	var __extAtomicExpression = '(?:'+typeSelector+'|'+cssClass+'|'+id+'|'+attribute+'|'+__simplePseudoFunctionnal+'|'+__atomicExpression+')';
	var __extArgStack = __extAtomicExpression+'(?:\\s*,\\s*'+__extAtomicExpression+')*';
// ######## Pseudo functionnal notations - note we don't use it in fact... we got an explicit list
	var pseudo = ':('+pseudoList+')(?:\\(\\s*('+__extArgStack+')?\\s*\\))?';

// These are the only one used by the parser
	gEBS_cssSelectorParser.sepGrabber = '^(?:\\s*('+combinators+')\\s*|(\\s*,\\s*))';
	gEBS_cssSelectorParser.tagNsGrabber = '^'+typeSelector;
	gEBS_cssSelectorParser.simpleSelGrabber = '^(?:'+id+'|'+cssClass+'|'+attribute+'|'+pseudo+')';
};

gEBS_cssSelectorParser.onload();

/**
 * Will split any group in an Array of "thingies" representing groups:
 * Conformance:
 * a group of selectors containing an invalid selector is invalid.
 * a selector containing an invalid simple selector, an invalid combinator or an invalid token is invalid.
 * a simple selector containing an undeclared namespace prefix is invalid
 */

gEBS_cssSelectorParser.loadGroup = function(selectorGroup){
// Will hold a list of selectors data
	this.group = [];
// Prepare the first slot for the first selector
	var x = 0;
	var weight = 0;
	this.group[x] = [];

// For debugin
	var sourceGroup = selectorGroup;
// Initiate that
	var previous = 0;
// Loop
	while(selectorGroup.length && previous != selectorGroup.length){
	// Set it. If we don't match, we'll get out of the loop and spit an error.
		previous = selectorGroup.length;
	// First try to grab something in a sequence
		var returnSelectorGroup = this.loadSequence(selectorGroup);
		var sequenceData = returnSelectorGroup.shift();
		returnSelectorGroup = returnSelectorGroup.shift();
	// If we failed parsing, loop will end here, and it will get invalidated outer
		if ((returnSelectorGroup == selectorGroup) || returnSelectorGroup == '##dropit'){
			gEBS_printfire('Error while parsing selector sequence');
		}else{
		// Set the new rest and new previous
			selectorGroup = returnSelectorGroup;
			previous = selectorGroup.length;
		// Push the valid data we got
			weight += sequenceData.weight;
			this.group[x].push(sequenceData);
		// If we still have something, it MUST be either a comma or combi
			if(selectorGroup.length){
			 	var reg = new RegExp(this.sepGrabber, 'i');
				var tmp = selectorGroup.match(reg);
			// If we have nothing, we'll get out of the loop
				if(tmp){
				// Got something
					selectorGroup = selectorGroup.replace(reg, '');
					tmp.shift();
					var comma = tmp.pop();
					var combi = tmp.pop();
				// In the middle of a sequence? Push the combi, move on.
					if(combi){
						this.group[x].push(combi);
				// At the end of a sequence? Push the weight. Increment.
				// Initiate the next container
					}else if(comma){
						this.group[x].push(weight);
						x++;
						weight = 0;
						this.group[x] = [];
					}
				}
			}
		}
	}
// Don't allow that..
	if(!this.group[x].length){
		gEBS_printfire('Invalid group:'+sourceGroup);
		selectorGroup = '##dropit';
	}
// Push the last weight on the last stack
	else{
		this.group[x].push(weight);
	}
// If we still got something at the end of the group, that's just bad
	if(selectorGroup){
		gEBS_printfire('Unparsable data in a group - group discarded:'+sourceGroup);
		selectorGroup = '##dropit';
	}
	return selectorGroup;
};


/**
 * Will split-up a selector sequence into components
 */

gEBS_cssSelectorParser.loadSequence = function(selectorSequence, denyNot){
// Initiate that
	var sequence = {
		selected: false,
		tagName: '*',
		nameSpace: '*',
		ids: [],
		classes: [],
		attributes: [],
		pseudoClasses: [],
		pseudoElements: [],
		weight: 0
	};

// Grab a potential selected bang
    if(selectorSequence.charAt(0) == '!'){
		sequence.selected = true;
		selectorSequence = selectorSequence.substr(1);
    }

// Grab a potential NS|tagName...
	var reg = new RegExp(this.tagNsGrabber, 'i');
	var match = selectorSequence.match(reg);
	if(match){
	// Split it up
		match = match[0].split('|');
		sequence.tagName = match.pop();
		NS = match.pop();
	// Weight it
		sequence.weight = (sequence.tagName != '*') ? 1 : 0;
	// no NS specified: if we got a default NS, it will be used, if not, we will match as *
	// FIXME: IE will break here most probably
		if(typeof NS == 'undefined'){
			if((defaultNs = this.getNameSpace()))
				sequence.nameSpace = defaultNs;
			else
				sequence.nameSpace = '*';
	// Now, if we got a namespace prefix, that is NOT *
		}else if(NS && NS != '*'){
		// Prefix not declared => drop the whole shit
			if(!(thatNSUri = this.getNameSpace(NS))){
				gEBS_printfire('Use of undeclared namespace in tag name.\nRemaining unparsed data:' + selectorSequence);
				selectorSequence = '##dropit';
		// Else, map the prefix to the uri
			}else{
				sequence.nameSpace = thatNSUri;
			}
	// null NS means we will match elements with NO namespace ONLY, so NS goes null
		}else if(NS != '*'){
			sequence.nameSpace = '';
		}

	// Strip it from the selector
		selectorSequence = selectorSequence.replace(reg, '');
	}

// Now, enter the loop until we can't match no more
	var previous = 0;
	var reg = new RegExp(this.simpleSelGrabber, 'i');
	while(selectorSequence.length && (selectorSequence != '##dropit') && selectorSequence.length != previous){
		var previous = selectorSequence.length;
		var match = selectorSequence.match(reg);
		if(match){
			match.shift();
			if((matchId = match.shift())){
				sequence.weight += 10000;
				sequence.ids.push(matchId);
			}else if((matchClass = match.shift())){
				sequence.weight += 100;
				sequence.classes.push(matchClass);
			}else if((matchAttrName = match.shift()) || match.shift() || match.shift()){
				sequence.weight += 100;
				var matchAttrRel = match.shift();
				var matchAttrValue = match.shift();
				if(matchAttrValue && (matchAttrValue.charAt(0) == '"' || matchAttrValue.charAt(0) == "'"))
					matchAttrValue = matchAttrValue.substr(1, matchAttrValue.length-2);
				matchAttrName = matchAttrName.split('|');
				var attrName = matchAttrName.pop();
				var attrNS = matchAttrName.pop();
// FIXME: IE might break here
				if(typeof attrNS == 'undefined')
					attrNS = '';
				if(attrNS && attrNS != '*')
					if(!(thatNSUri = this.getNameSpace(attrNS))){
						gEBS_printfire('Use of undeclared namespace in attribute name.\nRemaining unparsed data:' + selectorSequence);
						selectorSequence = '##dropit';
					}else{
						attrNS = thatNSUri;
					}
				sequence.attributes.push(Array(attrNS, attrName, matchAttrRel, matchAttrValue));
			}
			else if((matchFunk = match.shift()))
			{
				var matchArgs = match.shift();
				var matchFunk = gEBS_mapCssName2Js(matchFunk);
// FIXME: validate this
				if(matchFunk.charAt(0) == ':')
				{
					matchFunk = matchFunk.substr(1);
					sequence.pseudoElements.push(matchFunk);
				}
				else
				{
				// Not is anyhow a special case we MUST handle specifically...
				// Rats it!
				// Note to self: flush out the dynamics from there, and map them 
				// to new pseudo dynamics (eg: :not-hover)
				// Note to self: decide what to do with pseudo elements
					if(matchFunk == 'not'){
						if(denyNot){
							gEBS_printfire('You can\'t use a :not in a :not.\nRemaining unparsed data:' + selectorSequence);
							selectorSequence = '##dropit';
// FIXME: check IE
						}else if(typeof matchArgs == 'undefined'){
							gEBS_printfire('Unrecognised data sequence in :not argument.\nRemaining unparsed data:' + selectorSequence);
							selectorSequence = '##dropit';
						}else{
							matchArgs = this.loadSequence(matchArgs, 'denyNots');
							unparsedData = matchArgs.pop();
							matchArgs = matchArgs.pop();
							sequence.weight += matchArgs.weight;
						// No "selected bang" allowed, of course...
							matchArgs.selected = '';
							if (unparsedData){
								selectorSequence = '##dropit';
							}
						}
					}else{
				// Note to self: parse the argument stack in some way to help later?
						sequence.weight += 100;
					}
					sequence.pseudoClasses.push(Array(matchFunk, matchArgs));
				}
			}
			selectorSequence = selectorSequence.replace(reg, '');
		}
	}
	return Array(sequence, selectorSequence);
};


// src/parser/namespaces.js",

/**
 * This to handle nameSpaces declaration / retrieval in the selector parser
 */


gEBS_cssSelectorParser.spaces = {};
gEBS_cssSelectorParser.defaultNS = '';
gEBS_cssSelectorParser.addNameSpace = function(nameSpace, prefix){
	if(!prefix)
		this.defaultNS = nameSpace;
	else
		this.spaces[prefix] = nameSpace;
};

gEBS_cssSelectorParser.getNameSpace = function(prefix){
	if(!prefix)
		return this.defaultNS;
	else if(typeof this.spaces[prefix] != 'undefined')
		return this.spaces[prefix];
	else
		return '';
};

gEBS_cssSelectorParser.clearNameSpaces = function(){
	this.spaces = {};
	this.defaultNS = '';
};



// src/checkers/core.js",

/**
 * The whole object
 */

var gEBS_cssCheckers = {};

/**
 * This handle a complete sequence on an element
 */

gEBS_cssCheckers.nameAndNS = function(element, ns, name){
	var exit = true;
	// Brain Dead bro goes here
	if(typeof element.hasTagNameNS != 'undefined'){
		exit = element.hasTagNameNS(ns, name);	
	// Others go here
	}else{
		if(name != '*')
// FIXME: case insensitive
			exit = (element.nodeName.toLowerCase() == name.toLowerCase());
		if(ns == ''){
			exit = exit && (!element.namespaceURI);
		}else if(ns != '*'){
			exit = exit && (element.namespaceURI == ns);
		}
	}
	return exit;
};

gEBS_cssCheckers.sequence = function(element, splittedSequence){
    var exit = true;
// Tag & NS: MUST have been already resolved and mapped to MANDATORY values
	exit = this.nameAndNS(element, splittedSequence.nameSpace, splittedSequence.tagName);
	var x = 0;
// FIXME: xml:id and consorts...
	while(exit && x < splittedSequence.ids.length){
		exit = this.attribute(element, '*', 'id', '=', splittedSequence.ids[x]);
		x++;
	}
// FIXME? NS for class?
// FIXME: maybe replace classes by [*|class~="value"] in order to simplify here a bit?
	var x = 0;
	while(exit && x < splittedSequence.classes.length){
		exit = this.attribute(element, '*', 'class', '~=', splittedSequence.classes[x]);
		x++;
	}
// Attributs
	var x = 0;
	while(exit && x < splittedSequence.attributes.length){
		exit = this.attribute(element, splittedSequence.attributes[x][0], splittedSequence.attributes[x][1], splittedSequence.attributes[x][2], splittedSequence.attributes[x][3]);
		x++;
	}
// Pseudo classes (funkName / args)
	var x = 0;
	while(exit && x < splittedSequence.pseudoClasses.length){
	// Charge au fonctionnel de renvoyer -1 si les arguments sont invalides
	// Not est sp?cial...
		exit = gEBS_cssHandlers.pseudoClassesHandler(element, splittedSequence.pseudoClasses[x][0], splittedSequence.pseudoClasses[x][1]);
		x++;
	}
// FIXME Pseudo Elements... do something for christ sake!!!
	return exit;
};


/**
 * This is just a helper for the sequence matcher above
 * Attribute matching Handler
 * Honors any namespace fancyness
 */

gEBS_cssCheckers.attribute = function(element, nameSpace, attrName, matchingMethod, valueToMatch)
{
	var exit = false;
	if(nameSpace == '*'){
		var x = 0;
		var list = [];
		var reg = new RegExp('^(?:[^:]+:)?'+gEBS_regEscape(attrName)+'$', 'i');
		while(x < element.attributes.length){
			if(element.attributes[x].nodeName.match(reg)){
				exit = true;
				list.push(element.attributes[x]);
			}
			x++;
		}
		if(exit && matchingMethod){
			exit = false;
			while(list.length && !exit){
				var thatCandidate = list.pop().nodeValue;
				exit = gEBS_cssHandlers.matchersHandler(valueToMatch, matchingMethod, thatCandidate);
			}
		}
	}else{
		var exit = element.hasAttributeNS(nameSpace, attrName);
	    if(exit && matchingMethod){
	    	var actualValue = element.getAttributeNS(nameSpace, attrName);
			exit = gEBS_cssHandlers.matchersHandler(valueToMatch, matchingMethod, actualValue);
	    }
	}
	return exit;
};







// src/core.js"


// MatchUp being an array of (seq, combi, seq, .., seq) to be applied on "us+parents" (combinator wise)
function cssMatchUp(elt, matchUp, upLimitElement){
// Get what we should match
	var mySelectorSeq = matchUp.pop();
// Prepare the exit
	var finalResult = false;
// Validate me
	var meGood = gEBS_cssCheckers.sequence(elt, mySelectorSeq);
	if(meGood){
	// If on top of the mountain, upLimitElement MUST be in our parents
	    if(!matchUp.length){
			var candide = elt;
	    	var parents = [];
			while(candide.parentNode){
				candide = candide.parentNode;
				parents.push(candide);
			}
			if(gEBS_arrayContains(parents, upLimitElement)){
	    		finalResult = mySelectorSeq.selected ? Array(elt) : true;
			}
	// Guy! Still to climb and already on top of the mountain??? Dish that!!!
		}else if(elt == upLimitElement){
			meGood = false;
	// Ok, let pass the shit upstairs
		}else{
	    // Up combinator...
	        var upCombinator = matchUp.pop();
		// a - Grab candidates according to the combinator
			var upCandidates = gEBS_cssHandlers.combinatorsHandler(elt, upCombinator, 'up');
		// b - Grab upstairs, until the end of candides, or we validate (once is enough if no selected bang upstairs)
		// if selection there is, we'll go to the end of the cans
			while(upCandidates.length && (finalResult == false || typeof finalResult == 'object')){
				var thatCandidate = upCandidates.pop();
			// Grab upstairs
	            var grannyAndGranPa = cssMatchUp(thatCandidate, matchUp, upLimitElement);
			// What's in there? A collection of selected elements? Flush' em
				if(typeof grannyAndGranPa == 'object'){
					if(typeof finalResult != 'object')
						finalResult = [];
		            for(var y = 0; y < grannyAndGranPa.length; y++)
		                if(!gEBS_arrayContains(finalResult, grannyAndGranPa[y]))
		                    finalResult.push(grannyAndGranPa[y]);
			// Else, it's either false or true...
				}else{
					finalResult = grannyAndGranPa;
				}
			}
		// c - fine. Do we got a yes without selected elements? If so, we are the answer, whatever we are...
			if(finalResult && typeof finalResult != 'object')
	    		finalResult = mySelectorSeq.selected ? Array(elt) : true;
		// d - Oh god I hate javascript! *$??$$****
			matchUp.push(upCombinator);
		}
	}
// Repush that...
	matchUp.push(mySelectorSeq);
// Return nicely...
	return finalResult;
};

// MatchDown being an array of (selector, combinator, selector), the first being applied on this
function cssMatchDown(elt, matchDown){
// Get the selecterSequence for the current element
	var mySelectorSeq = matchDown.shift();
// Prepare the exit...
	var finalResult = [];
// Check it's ok (dead selector means dirty panda hacketry, so true)
	var meGood = (typeof mySelectorSeq.selected != 'undefined') ? gEBS_cssCheckers.sequence(elt, mySelectorSeq) : true;
// If it is
	if(meGood){
	// If this is the end of the road... returns ourselves...
	// :selected() or not, we ARE the end of the road
	    if(!matchDown.length){
	        finalResult = Array(elt);
	// Else, well... go down...
	    }else{
		// a - Grab down according to the combinator
		// Note: we don't bother checking there's a function for the combi
		// If somebody did register an invalid combi, he drops dead
	        var downCombinator = matchDown.shift();
			var downCandidates = gEBS_cssHandlers.combinatorsHandler(this, downCombinator, 'down');
		// b - force them propagate nicely down
	        for(var x = 0; x < downCandidates.length; x++){
	            var tmp = cssMatchDown(downCandidates[x], matchDown);
	            for(var y = 0; y < tmp.length; y++)
	                if (!gEBS_arrayContains(finalResult, tmp[y]))
	                    finalResult.push(tmp[y]);
	        }
	    // c - if we are selected, we ARE the result if any result...
	    	if(finalResult.length && mySelectorSeq.selected)
	    		finalResult = Array(elt);
		// d - Oh god I hate javascript! *$??$$****
		    matchDown.unshift(downCombinator);
	    }
	}
    matchDown.unshift(mySelectorSeq);
	return finalResult;
};


function getElementsBySplittedSelector(elt, goUp){
	// Handle first-letter, first-line here...
	if(goUp && (typeof goUp[goUp.length-1].pseudoElements) && gEBS_arrayContains(goUp[goUp.length-1].pseudoElements, 'firstLetter')){
		firstLetter = true;
	}else{
		firstLetter = false;
	}
	if(goUp && (typeof goUp[goUp.length-1].pseudoElements) && gEBS_arrayContains(goUp[goUp.length-1].pseudoElements, 'firstLine')){
		firstLine = true;
	}else{
		firstLine = false;
	}

// Prepare the exit
    var globalResult = [];
// Will store combi, seq, combi, seq... to go "down" beginning with our start point
    var downStack = [];
// This will become an array if we find id / tagName on some fragment
    var startElements = false;
// In the case we found something, will contain effectively appropriate elements
// eg: elements that match the sequence AND which are ok regarding all their parents
    var realStart = [];
// In the case we have really matching elements, and if something is selected somewhere
// (either them or their parents), this will live in there
    var selectedElements = {};


// While the stack contains smthing, pop it up to find id / tagName...
	while(goUp.length && !startElements){
		var selectorSequence = goUp.pop();
	// Is there an id / tag in this fragment?
		if(selectorSequence.ids.length){
			tmp = document.getElementById(selectorSequence.ids[0]);
            startElements = (tmp ? Array(tmp) : Array());
		}else if(selectorSequence.tagName && selectorSequence.tagName != '*'){
            startElements = elt.getElementsByTagNameNS(selectorSequence.nameSpace, selectorSequence.tagName);
		}
		if(startElements){
		// Got something. Refill the stack.
			goUp.push(selectorSequence);
		// MatchDown is either empty or does begin with a combi.
		// Anyhow, we want to match ourselves without actually doing *again* anything
			downStack.unshift('');
			var goDown = downStack;
		}else{
		// Nothing. Simply shift the data and potential combi down the drain.
			downStack.unshift(selectorSequence);
			combinator = goUp.pop();
			if(combinator)
				downStack.unshift(combinator);
		}
	}

// Ok... If we didn't started already, then we need to start here. And by the way, goUp
    if(!startElements){
	// Beware of IE!!!!!!!!! It will return also comments in there!!!!!
    	startElements = elt.getElementsByTagNameNS('*', '*');
    	goUp = downStack;
	// We'll need that only to "validate"
    	goDown = false;
    }


// Validate the candidates, grab the selected and "realstart"...
    for(var y = 0; y < startElements.length; y++){
	// Take II: Beware of IE!!!!!!!!! It will return also comments in there!!!!!
		if(startElements[y].nodeType == 1){
	        var parents = cssMatchUp(startElements[y], goUp, elt);
	        // If we got true, or an Array of selected Elements.
			if(parents){
	            realStart.push(startElements[y]);
	            if(typeof parents == 'object')
	            	selectedElements[startElements[y]] = parents;
			}
		}
    }

// Now, we may have to go down (in case we started from an id/ns|tag point)
// So let's do it quick and painless

    for(var y = 0; y < realStart.length; y++){
    	if(goDown)
	        var localResult = cssMatchDown(realStart[y], goDown);
		else
			var localResult = [realStart[y]];
        if(localResult.length){
		// Ah! we were selecting already
        	if(typeof selectedElements[realStart[y]] != 'undefined'){
		        for(var x = 0; x < selectedElements[realStart[y]].length; x++){
		            if(!gEBS_arrayContains(globalResult, selectedElements[realStart[y]][x])){
		        		globalResult.push(selectedElements[realStart[y]][x]);
		            }
		        }
        	}else{
		// We were not. Simply flush what we grabbed, selection is done if any
		        for(var z = 0; z < localResult.length; z++)
		            if(!gEBS_arrayContains(globalResult, localResult[z]))
		                globalResult.push(localResult[z]);
        	}
        }
    }
    return globalResult;
};



/**
 * This function is a gluer between:
 * a) the selector's parser
 * b) the checker's functions
 * c) the element attached functions responsible for wandering through all this
 */

function getElementsBySelector(elt, selectorGroup, showTime){
// Split the supposed group
  var start = new Date();
	var tmp = gEBS_cssSelectorParser.loadGroup(selectorGroup);
  var end = new Date();
	if(showTime){
	    gEBS_printfire('getElementsBySelector called on group: '+selectorGroup);
	    gEBS_printfire('Selector parsing needed: '+(end-start)+' milliseconds');
	}
// If there's unparsed data, then the *whole* group is invalid. Drop it.
	if(tmp)
		gEBS_cssSelectorParser.group = [];
// Now, for each element (splitted selector) of the group, namely goUp
  var start = new Date();
	var globalResult = [];
	for(var x = 0; x < gEBS_cssSelectorParser.group.length && (goUp = gEBS_cssSelectorParser.group[x]); x++){
	// Just dish the weight for now...
		goUp.pop();
	// Give it as is to the manger
		var localResult = getElementsBySplittedSelector(elt, goUp);
		for(var y = 0; y < localResult.length; y++){
			if(!gEBS_arrayContains(globalResult, localResult[y]))
				globalResult.push(localResult[y]);
		}
	}
  var end = new Date();
	if(showTime){
		gEBS_printfire('Actual selection needed: '+(end-start)+' milliseconds');
	}
    return globalResult;
};

function matchesSelector(rootContext, selector, elementToMatch)
{
// Split the candidate
	var tmp = gEBS_cssSelectorParser.loadGroup(selector);
  var exit = false;
// If there's unparsed data, then its invalid. Otherwise, we search.
  if(!tmp)
  {
    // Don't deal with groups, just a selector
    var goUp = gEBS_cssSelectorParser.group[0];
    // And dish the weight for now...
    goUp.pop();
    // Grab the result in context
    var localResult = getElementsBySplittedSelector(rootContext, goUp);
    if(gEBS_arrayContains(localResult, elementToMatch))
      exit = true;
  }
  return exit;
};
