/*
MIT License
Copyright (c) 2006 Sam Ruby

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the “Software?), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED AS IS, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/

var MathML = {
  // Jacques Distler subset of Presentational MathML
  root: 'math',
  ns: 'http://www.w3.org/1998/Math/MathML',
  elements: ['maction', 'math', 'merror', 'mfrac', 'mi',
    'mmultiscripts', 'mn', 'mo', 'mover', 'mpadded', 'mphantom',
    'mprescripts', 'mroot', 'mrow', 'mspace', 'msqrt', 'mstyle', 'msub',
    'msubsup', 'msup', 'mtable', 'mtd', 'mtext', 'mtr', 'munder',
    'munderover', 'none'],
  attributes: ['actiontype', 'align', 'columnalign', 'columnalign',
    'columnalign', 'columnlines', 'columnspacing', 'columnspan', 'depth',
    'display', 'displaystyle', 'equalcolumns', 'equalrows', 'fence',
    'fontstyle', 'fontweight', 'frame', 'height', 'linethickness', 'lspace',
    'mathbackground', 'mathcolor', 'mathvariant', 'mathvariant', 'maxsize',
    'minsize', 'other', 'rowalign', 'rowalign', 'rowalign', 'rowlines',
    'rowspacing', 'rowspan', 'rspace', 'scriptlevel', 'selection',
    'separator', 'stretchy', 'width', 'width']
}

var SVG = {
  // svgtiny + class + opacity + offset + style
  root: 'svg',
  ns: 'http://www.w3.org/2000/svg',
  elements: ['a', 'animate', 'animateColor', 'animateMotion',
    'animateTransform', 'circle', 'defs', 'desc', 'ellipse', 'font-face',
    'font-face-name', 'font-face-src', 'g', 'glyph', 'hkern', 'image',
    'linearGradient', 'line', 'metadata', 'missing-glyph', 'mpath', 'path',
    'polygon', 'polyline', 'radialGradient', 'rect', 'set', 'stop',
    'switch', 'text', 'title', 'use'],
  attributes: ['accent-height', 'accumulate', 'additive', 'alphabetic',
    'arabic-form', 'ascent', 'attributeName', 'attributeType',
    'baseProfile', 'bbox', 'begin', 'by', 'calcMode', 'cap-height',
    'class', 'color', 'color-rendering', 'content', 'cx', 'cy', 'd',
    'descent', 'display', 'dur', 'end', 'fill', 'fill-rule', 'font-family',
    'font-size', 'font-stretch', 'font-style', 'font-variant',
    'font-weight', 'from', 'fx', 'fy', 'g1', 'g2', 'glyph-name', 'hanging',
    'height', 'horiz-adv-x', 'horiz-origin-x', 'id', 'ideographic', 'k',
    'keyPoints', 'keySplines', 'keyTimes', 'lang', 'mathematical', 'max',
    'min', 'name', 'offset', 'opacity', 'origin', 'overline-position',
    'overline-thickness', 'panose-1', 'path', 'pathLength', 'points',
    'preserveAspectRatio', 'r', 'repeatCount', 'repeatDur',
    'requiredExtensions', 'requiredFeatures', 'restart', 'rotate', 'rx',
    'ry', 'slope', 'stemh', 'stemv', 'stop-color', 'stop-opacity',
    'strikethrough-position', 'strikethrough-thickness', 'stroke',
    'stroke-dasharray', 'stroke-dashoffset', 'stroke-linecap',
    'stroke-linejoin', 'stroke-miterlimit', 'stroke-width', 'style',
    'systemLanguage', 'target', 'text-anchor', 'to', 'transform', 'type',
    'u1', 'u2', 'underline-position', 'underline-thickness', 'unicode',
    'unicode-range', 'units-per-em', 'values', 'version', 'viewBox',
    'visibility', 'width', 'widths', 'x', 'x-height', 'x1', 'x2',
    'y', 'y1', 'y2', 'zoomAndPan']
}

  // clone a DOM subtree
  function _FS_deepcopy(aDoc, module, source, dest, nsmap) {
    // copy attributes
    for (var i=0; i<source.attributes.length; i++) {
      var oldattr = source.attributes[i];
      var colon = oldattr.name.indexOf(':');
      if (colon == -1) {
        for (var j=0; j<module.attributes.length; j++) {
          if (module.attributes[j].toLowerCase() != oldattr.name) continue;
          dest.setAttribute(module.attributes[j], oldattr.value);
          break;
        }
      } else { // namespace prefixed attribute
        var prefix = oldattr.name.slice(0,colon);
        var name = oldattr.name.slice(colon+1);
        if (prefix == 'xmlns') {
          var oldmap = nsmap;
          nsmap = {};
          for (var property in oldmap) nsmap[property] = oldmap[property];
          nsmap[name] = oldattr.value;
        } else {
          for (var ns in nsmap) {
            if (ns == prefix) {
              dest.setAttributeNS(nsmap[prefix], name, oldattr.value);
            }
          }
        }
      }
    }

    // copy children
    for (var i=0; i<source.childNodes.length; i++) {
      var oldchild = source.childNodes[i];
      if (oldchild.nodeType == 1) { // element
        for (var j=0; j<module.elements.length; j++) {
          if (module.elements[j].toUpperCase() != oldchild.nodeName) continue;
          var newchild = aDoc.createElementNS(module.ns,module.elements[j]);
          _FS_deepcopy(aDoc, module, oldchild, newchild, nsmap);
          dest.appendChild(newchild);
          break;
        }
      } else if (oldchild.nodeType == 3) { // text
        var newchild = aDoc.createTextNode(oldchild.nodeValue);
        dest.appendChild(newchild);
      }
    }
  }

function EnableSVGInHTML4(aDocument)
{

  // let's check we're in HTML 4...
  if (aDocument &&
      aDocument.documentElement.nodeName.toLowerCase() == "html" &&
      (!aDocument.doctype || aDocument.doctype.publicId.indexOf("HTML 4") != -1))
  {
    // copy modules into their appropriate namespaces
    var modules = [MathML, SVG];
    for (var i=0; i<modules.length; i++) {
      var module = modules[i];
      var roots = aDocument.getElementsByTagName(module.root);
      for (var j=0; j<roots.length; j++) {
        var source = roots[j];
        if (aDocument.createElementNS) {
          // if (source.__proto__ != HTMLUnknownElement.prototype) continue;
          if (source.getAttribute('xmlns') != module.ns) continue;
          var dest = aDocument.createElementNS(module.ns, module.root);
          _FS_deepcopy(aDocument, module, source, dest, {});
          source.parentNode.insertBefore(dest, source);
          source.parentNode.removeChild(source);
        } else {
          // fallback browsers that don't support DOM namespaces
          var img = aDocument.createElement('img');
          img.src = module.root + ' image';
          img.title = module.root + ' image';
          source.parentNode.insertBefore(img, source);
        }
      }
    }
  }
}
